#include <math.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <time.h>

//#include <iostream.h>
#include <fstream.h>

#include "emel.h"		
#include "kinematics.h"
#include "peakmap.h"
#include "mcdefs.h"


unsigned PART = ISR;
	
//unsigned NUMCOUNT = 500000;
unsigned MINIT = 2;
unsigned MAXIT = 2;
double CHI2LIM = 1.0;
//double SQRTS = 500.;

unsigned _VeryVerbose = 4;
unsigned long cutcount = 0, mapmcount = 0, totalcount = 0;
	

void randomize(void)
{
  srand((unsigned)time(NULL));
}

void random(double *r)
{
  *r = ((double)rand())/RAND_MAX;
}


//double PHOTENCUT = 1.0;
#define COLLCUT    0.9962    // cut on collinear emmission of photon,
			      //  cos(5) = 0.9962

// NOTE: costheta is only integrated from 0 to 1 since the process it
//  FB-symmetric. Therefore cuts have to be applied to one half-sphere only.

int cut(double s, cvector4 p1, cvector4 p2,
		  cvector4 k1, cvector4 k2, cvector4 k3)
{
//  return 0;
  return (k3[0] < PHOTENCUT) /*||
//	 (abs(Dot(p1,k3)/(p1[0]*k3[0]) -1) > COLLCUT) ||
//	 (abs(Dot(p2,k3)/(p2[0]*k3[0]) -1) > COLLCUT)*/;
}


static double rran, rmin, /*tran,*/ ME2;

void initmapvals(double s)
{
  double ssq = sqrt(s);
  
  ME2 = ME*ME;
//  tran = log(s/(2*ME2*(1+ME2/s)));
  rran = log((1 - 4*msmr*msmr/s)*ssq/(2*PHOTENCUT));
  rmin = log(2*PHOTENCUT/ssq);
}

double MapoooColl(double *xran, double s, double ssq,
        double m1, vector4 k1, double m2, vector4 k2, double m3, vector4 k3,
  	double& _expt, double& _k30)
{
  static double q1[4], r1[4], r2[4];
  double fac, s1sq;
  double /*tmin,*/ k30, t, r, costheta;
  
  r = rran * xran[0] + rmin;
  k30 = ssq/2 * exp(r);
  
  s1sq = sqrt(s - 2*ssq*k30);
  fac = ssq/M_PI * rran * k30;
  costheta = xran[1];
  _expt = k30*ssq * (1- costheta * (1- (2*ME2/s)*(1+ME2/s)));
  fac *= 2./_expt;

  _TwoDecay(ssq, s1sq, 0.0, costheta, xran[2]*2*M_PI, q1, k3);
  	// don't take jacobian factor from _TwoDecay since this sometimes
  	// return nan due to numerical errors, instead the exact contribution
  	// k30/ssq is explicitly plugged in below
  fac *= 1/(2*s) * 2*M_PI * 1/(16*M_PI*M_PI) * k30/ssq
                 * SubDec(xran+3, s1sq, m1, r1, m2, r2);
  Boost(r1, q1, k1);
  Boost(r2, q1, k2);
  fac *= 2;	           // since we only integrate over one half-sphere
  
  _k30 = k30;
  return fac;
}  

double MapoooCollSubtr(double *xran, double s, double ssq,
        double m1, vector4 k1, double m2, vector4 k2, double m3, vector4 k3,
  	double& _expt, double& _k30)
{
  static double q1[4], r1[4], r2[4];
  double fac, s1sq;
  double /*tmin,*/ k30, t, r, costheta;
  
  r = rran * xran[0] + rmin;
  k30 = ssq/2 * exp(r);
  
  s1sq = sqrt(s - 2*ssq*k30);
  fac = ssq/M_PI * rran * k30;
  costheta = xran[1];
  _expt = k30*ssq * (1- costheta * (1- (2*ME2/s)*(1+ME2/s)));
  fac *= 2./_expt;
  
  _TwoDecay(ssq, s1sq, 0.0, 1.0, xran[2]*2*M_PI, q1, k3);
  	// don't take jacobian factor from _TwoDecay since this sometimes
  	// return nan due to numerical errors, instead the exact contribution
  	// k30/ssq is explicitly plugged in below
  fac *= 1/(2*s) * 2*M_PI * 1/(16*M_PI*M_PI) * k30/ssq
                 * SubDec(xran+3, s1sq, m1, r1, m2, r2);
  Boost(r1, q1, k1);
  Boost(r2, q1, k2);
  fac *= 2;	           // since we only integrate over one half-sphere
  
  _k30 = k30;
  return fac;
}  

double MapoooEg(double *xran, double s, double ssq,
        double m1, vector4 k1, double m2, vector4 k2, double m3, vector4 k3,
  	double& _k30)
{
  static double q1[4], r1[4], r2[4];
  double fac, s1sq;
  double k30, r, costheta;
  
  r = rran * xran[0] + rmin;
  k30 = ssq/2 * exp(r);
  s1sq = sqrt(s - 2*ssq*k30);
  fac = ssq/M_PI * rran * k30;

  fac *= 1/(2*s) * SubDec(xran+1, ssq, s1sq, q1, m3, k3)
                 * SubDec(xran+3, s1sq, m1, r1, m2, r2);
  Boost(r1, q1, k1);
  Boost(r2, q1, k2);
  _k30 = k30;
  return fac;
}  

#define NDIM	5

static double k1[4], k2[4], k3[4];

emerec evalMatrixEl(unsigned part, double s, double ctmin, double ctmax,
		    cvector4 p1, cvector4 p2)
{
  static double xran[NDIM];
  double ginv;
  double sqrts = sqrt(s), ct, expt, Eg, r;
  unsigned j;
  emerec rtn;
  
  for(j = 0; j < NDIM; j++)
    random(&xran[j]);
  
  ginv = MapoooColl(xran, s, sqrts, msmr, k1, msmr, k2, 0.0, k3, expt, Eg);
  if(ginv == 0.0)
  {
    rtn.state = MAPMISS;
    return rtn;
  }
  ct = (p1[1]*k1[1] + p1[2]*k1[2] + p1[3]*k1[3]) /
	  (p1[0] * sqrt(k1[1]*k1[1] + k1[2]*k1[2] + k1[3]*k1[3]));
  if(ct < ctmin || ct > ctmax)
  {
    rtn.state = MAPMISS;
    return rtn;
  }
  
  if(cut(s, p1, p2, k1, k2, k3))
  {
    rtn.state = CUTHIT;
    return rtn;
  }

  rtn.state = NORMAL;
  rtn.val = 0.0;
  if((part & ISR) != 0)
  {
#include "crss0.isr.num"
    rtn.val += r;
  }

  if((part & ISR) != 0)
  {
    ginv = MapoooCollSubtr(xran, s, sqrts, msmr, k1, msmr, k2, 0.0, k3, expt, Eg);
#include "crss0.isrfac.num"
    rtn.val -= r;
  }
  
  rtn.val *= ginv;
  rtn.dnsty = 1/ginv;
  return rtn;
}


void printtime(void)
{
  static clock_t lasttime = -1, ptime;

  ptime = clock() / CLOCKS_PER_SEC;
  if(lasttime >= 0)
    printf(
     "     iteration time: %02u:%02u    total elapsed time: %02u:%02u:%02u\n",
	  (unsigned)((ptime-lasttime)/60),
	  (unsigned)(ptime-lasttime) % 60,
	  (unsigned)(ptime/3600),
	  (unsigned)(ptime/60) % 60,
	  (unsigned)ptime % 60);
  lasttime = ptime;
}


mocres evalMCpoint(int PART, double s, double ctmin, double ctmax,
		   unsigned NUMCOUNT)
{
  double p1[4], p2[4];
  emerec eval;
  double wval;
  double In, Wsq, Sd, Var, sum_all, sdi_all, s2_all, chi2a;
  unsigned long it, icount;
  mocres res;
  
  p1[0] = p2[0] = sqrt(s)/2;
//  p1[3] = -(p2[3] = sqrt(s)/2);
  p1[3] = -(p2[3] = sqrt(s/4-ME2));
  p1[1] = p1[2] = p2[1] = p2[2] = 0.;
  initmapvals(s);

  sum_all = sdi_all = s2_all = chi2a = 0.0;  

  for(it = 1; it <= MINIT || (it <= MAXIT && chi2a > CHI2LIM); it++)
  {
    In = Wsq = 0.0;
    for(icount = 0; icount < NUMCOUNT; icount++)
    {  
      eval = evalMatrixEl(PART, s, ctmin, ctmax, p1, p2);
      if(eval.state == MAPMISS)
      {
	mapmcount++;
	continue;
      }
      totalcount++;
      if(eval.state == CUTHIT)
      {
	cutcount++;
	continue;
      }

      wval = eval.val * TOFB;
      In += wval;
      Wsq += wval*wval;
    }
    In /= NUMCOUNT;
    Wsq /= NUMCOUNT;
    Var = (Wsq-In*In)/(NUMCOUNT-1);
    Sd = sqrt(Var);
    
    sum_all += In/Var;
    sdi_all += 1/Var;
    s2_all += In*In/Var;
    if(it > 1)
      chi2a = 1.0/(it-1) * (s2_all - sum_all*sum_all/sdi_all);
    else
      chi2a = 0.0;
    printf("#%2u  int: %8.4lf, %8.4lf  sum: %8.4lf, %8.4lf chi2: %8.4lf\n",
		it, In, Sd, sum_all/sdi_all, sqrt(1/sdi_all), chi2a);
    printtime();
  }
  res.avgi = sum_all/sdi_all;
  res.sd = sqrt(1/sdi_all);

  printf("res = %2.4lf, sd = %2.4lf\n", res.avgi, res.sd);
  return res;
}
